# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import testtools

from openstack.block_storage.v3 import volume
import openstack.cloud
from openstack.cloud import meta
from openstack.compute.v2 import volume_attachment
from openstack.tests import fakes
from openstack.tests.unit import base


class TestVolume(base.TestCase):

    def _compare_volumes(self, exp, real):
        self.assertDictEqual(
            volume.Volume(**exp).to_dict(computed=False),
            real.to_dict(computed=False)
        )

    def _compare_volume_attachments(self, exp, real):
        self.assertDictEqual(
            volume_attachment.VolumeAttachment(**exp).to_dict(computed=False),
            real.to_dict(computed=False)
        )

    def test_attach_volume(self):
        server = dict(id='server001')
        vol = {'id': 'volume001', 'status': 'available',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        rattach = {'server_id': server['id'], 'device': 'device001',
                   'volumeId': volume['id'], 'id': 'attachmentId'}
        self.register_uris([
            self.get_nova_discovery_mock_dict(),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'compute', 'public',
                     append=['servers', server['id'],
                             'os-volume_attachments']),
                 json={'volumeAttachment': rattach},
                 validate=dict(json={
                     'volumeAttachment': {
                         'volumeId': vol['id']}})
                 )])
        ret = self.cloud.attach_volume(server, volume, wait=False)
        self._compare_volume_attachments(rattach, ret)
        self.assert_calls()

    def test_attach_volume_exception(self):
        server = dict(id='server001')
        vol = {'id': 'volume001', 'status': 'available',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        self.register_uris([
            self.get_nova_discovery_mock_dict(),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'compute', 'public',
                     append=['servers', server['id'],
                             'os-volume_attachments']),
                 status_code=404,
                 validate=dict(json={
                     'volumeAttachment': {
                         'volumeId': vol['id']}})
                 )])
        with testtools.ExpectedException(
            openstack.cloud.OpenStackCloudURINotFound
        ):
            self.cloud.attach_volume(server, volume, wait=False)
        self.assert_calls()

    def test_attach_volume_wait(self):
        server = dict(id='server001')
        vol = {'id': 'volume001', 'status': 'available',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        vol['attachments'] = [{'server_id': server['id'],
                               'device': 'device001'}]
        vol['status'] = 'in-use'
        attached_volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        rattach = {'server_id': server['id'], 'device': 'device001',
                   'volumeId': volume['id'], 'id': 'attachmentId'}
        self.register_uris([
            self.get_nova_discovery_mock_dict(),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'compute', 'public',
                     append=['servers', server['id'],
                             'os-volume_attachments']),
                 json={'volumeAttachment': rattach},
                 validate=dict(json={
                     'volumeAttachment': {
                         'volumeId': vol['id']}})),
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', vol['id']]),
                 json={'volume': volume}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', vol['id']]),
                 json={'volume': attached_volume})
        ])
        # defaults to wait=True
        ret = self.cloud.attach_volume(server, volume)
        self._compare_volume_attachments(rattach, ret)
        self.assert_calls()

    def test_attach_volume_wait_error(self):
        server = dict(id='server001')
        vol = {'id': 'volume001', 'status': 'available',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        vol['status'] = 'error'
        errored_volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        rattach = {'server_id': server['id'], 'device': 'device001',
                   'volumeId': volume['id'], 'id': 'attachmentId'}
        self.register_uris([
            self.get_nova_discovery_mock_dict(),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'compute', 'public',
                     append=['servers', server['id'],
                             'os-volume_attachments']),
                 json={'volumeAttachment': rattach},
                 validate=dict(json={
                     'volumeAttachment': {
                         'volumeId': vol['id']}})),
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume['id']]),
                 json={'volume': errored_volume}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume['id']]),
                 json={'volume': errored_volume})
        ])

        with testtools.ExpectedException(
            openstack.exceptions.ResourceFailure
        ):
            self.cloud.attach_volume(server, volume)
        self.assert_calls()

    def test_attach_volume_not_available(self):
        server = dict(id='server001')
        volume = dict(id='volume001', status='error', attachments=[])

        with testtools.ExpectedException(
            openstack.cloud.OpenStackCloudException,
            "Volume %s is not available. Status is '%s'" % (
                volume['id'], volume['status'])
        ):
            self.cloud.attach_volume(server, volume)
        self.assertEqual(0, len(self.adapter.request_history))

    def test_attach_volume_already_attached(self):
        device_id = 'device001'
        server = dict(id='server001')
        volume = dict(id='volume001',
                      attachments=[
                          {'server_id': 'server001', 'device': device_id}
                      ])

        with testtools.ExpectedException(
            openstack.cloud.OpenStackCloudException,
            "Volume %s already attached to server %s on device %s" % (
                volume['id'], server['id'], device_id)
        ):
            self.cloud.attach_volume(server, volume)
        self.assertEqual(0, len(self.adapter.request_history))

    def test_detach_volume(self):
        server = dict(id='server001')
        volume = dict(id='volume001',
                      attachments=[
                          {'server_id': 'server001', 'device': 'device001'}
                      ])
        self.register_uris([
            self.get_nova_discovery_mock_dict(),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'compute', 'public',
                     append=['servers', server['id'],
                             'os-volume_attachments', volume['id']]))])
        self.cloud.detach_volume(server, volume, wait=False)
        self.assert_calls()

    def test_detach_volume_exception(self):
        server = dict(id='server001')
        volume = dict(id='volume001',
                      attachments=[
                          {'server_id': 'server001', 'device': 'device001'}
                      ])
        self.register_uris([
            self.get_nova_discovery_mock_dict(),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'compute', 'public',
                     append=['servers', server['id'],
                             'os-volume_attachments', volume['id']]),
                 status_code=404)])
        with testtools.ExpectedException(
            openstack.cloud.OpenStackCloudURINotFound
        ):
            self.cloud.detach_volume(server, volume, wait=False)
        self.assert_calls()

    def test_detach_volume_wait(self):
        server = dict(id='server001')
        attachments = [{'server_id': 'server001', 'device': 'device001'}]
        vol = {'id': 'volume001', 'status': 'attached', 'name': '',
               'attachments': attachments}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        vol['status'] = 'available'
        vol['attachments'] = []
        avail_volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        self.register_uris([
            self.get_nova_discovery_mock_dict(),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'compute', 'public',
                     append=['servers', server['id'],
                             'os-volume_attachments', volume.id])),
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', 'detail']),
                 json={'volumes': [avail_volume]})])
        self.cloud.detach_volume(server, volume)
        self.assert_calls()

    def test_detach_volume_wait_error(self):
        server = dict(id='server001')
        attachments = [{'server_id': 'server001', 'device': 'device001'}]
        vol = {'id': 'volume001', 'status': 'attached', 'name': '',
               'attachments': attachments}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        vol['status'] = 'error'
        vol['attachments'] = []
        errored_volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        self.register_uris([
            self.get_nova_discovery_mock_dict(),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'compute', 'public',
                     append=['servers', server['id'],
                             'os-volume_attachments', volume.id])),
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', 'detail']),
                 json={'volumes': [errored_volume]}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public',
                     append=['volumes', errored_volume['id']]),
                 json={'volume': errored_volume})
        ])
        with testtools.ExpectedException(
            openstack.exceptions.ResourceFailure
        ):
            self.cloud.detach_volume(server, volume)
        self.assert_calls()

    def test_delete_volume_deletes(self):
        vol = {'id': 'volume001', 'status': 'attached',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume.id]),
                 json={'volumes': [volume]}),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume.id])),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume.id]),
                 status_code=404)])
        self.assertTrue(self.cloud.delete_volume(volume['id']))
        self.assert_calls()

    def test_delete_volume_gone_away(self):
        vol = {'id': 'volume001', 'status': 'attached',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume.id]),
                 json=volume),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume.id]),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume.id]),
                 status_code=404),
        ])
        self.assertTrue(self.cloud.delete_volume(volume['id']))
        self.assert_calls()

    def test_delete_volume_force(self):
        vol = {'id': 'volume001', 'status': 'attached',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume['id']]),
                 json={'volumes': [volume]}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'volumev3', 'public',
                     append=['volumes', volume.id, 'action']),
                 validate=dict(
                     json={'os-force_delete': {}})),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', volume['id']]),
                 status_code=404)])
        self.assertTrue(self.cloud.delete_volume(volume['id'], force=True))
        self.assert_calls()

    def test_set_volume_bootable(self):
        vol = {'id': 'volume001', 'status': 'attached',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', 'detail']),
                 json={'volumes': [volume]}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'volumev3', 'public',
                     append=['volumes', volume.id, 'action']),
                 json={'os-set_bootable': {'bootable': True}}),
        ])
        self.cloud.set_volume_bootable(volume['id'])
        self.assert_calls()

    def test_set_volume_bootable_false(self):
        vol = {'id': 'volume001', 'status': 'attached',
               'name': '', 'attachments': []}
        volume = meta.obj_to_munch(fakes.FakeVolume(**vol))
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes', 'detail']),
                 json={'volumes': [volume]}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'volumev3', 'public',
                     append=['volumes', volume.id, 'action']),
                 json={'os-set_bootable': {'bootable': False}}),
        ])
        self.cloud.set_volume_bootable(volume['id'])
        self.assert_calls()

    def test_get_volume_by_id(self):
        vol1 = meta.obj_to_munch(fakes.FakeVolume('01', 'available', 'vol1'))
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'volumev3', 'public',
                     append=['volumes', '01']),
                 json={'volume': vol1}
                 )
        ])
        self._compare_volumes(vol1, self.cloud.get_volume_by_id('01'))
        self.assert_calls()

    def test_create_volume(self):
        vol1 = meta.obj_to_munch(fakes.FakeVolume('01', 'available', 'vol1'))
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes']),
                 json={'volume': vol1},
                 validate=dict(json={
                     'volume': {
                         'size': 50,
                         'name': 'vol1',
                     }})),
        ])

        self.cloud.create_volume(50, name='vol1')
        self.assert_calls()

    def test_create_bootable_volume(self):
        vol1 = meta.obj_to_munch(fakes.FakeVolume('01', 'available', 'vol1'))
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'volumev3', 'public', append=['volumes']),
                 json={'volume': vol1},
                 validate=dict(json={
                     'volume': {
                         'size': 50,
                         'name': 'vol1',
                     }})),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'volumev3', 'public',
                     append=['volumes', '01', 'action']),
                 validate=dict(
                     json={'os-set_bootable': {'bootable': True}})),
        ])

        self.cloud.create_volume(50, name='vol1', bootable=True)
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.clustering.v1 import policy
from openstack.tests.unit import base


FAKE_ID = 'ac5415bd-f522-4160-8be0-f8853e4bc332'
FAKE_NAME = 'test_policy'

FAKE = {
    'id': FAKE_ID,
    'name': FAKE_NAME,
    'spec': {
        'type': 'senlin.policy.deletion',
        'version': '1.0',
        'properties': {
            'criteria': 'OLDEST_FIRST',
            'grace_period': 60,
            'reduce_desired_capacity': False,
            'destroy_after_deletion': True,
        }
    },
    'project': '42d9e9663331431f97b75e25136307ff',
    'domain': '204ccccd267b40aea871750116b5b184',
    'user': '3747afc360b64702a53bdd64dc1b8976',
    'type': 'senlin.policy.deletion-1.0',
    'created_at': '2015-10-10T12:46:36.000000',
    'updated_at': '2016-10-10T12:46:36.000000',
    'data': {},
}


class TestPolicy(base.TestCase):

    def setUp(self):
        super(TestPolicy, self).setUp()

    def test_basic(self):
        sot = policy.Policy()
        self.assertEqual('policy', sot.resource_key)
        self.assertEqual('policies', sot.resources_key)
        self.assertEqual('/policies', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_instantiate(self):
        sot = policy.Policy(**FAKE)
        self.assertEqual(FAKE['id'], sot.id)
        self.assertEqual(FAKE['name'], sot.name)
        self.assertEqual(FAKE['spec'], sot.spec)
        self.assertEqual(FAKE['project'], sot.project_id)
        self.assertEqual(FAKE['domain'], sot.domain_id)
        self.assertEqual(FAKE['user'], sot.user_id)
        self.assertEqual(FAKE['data'], sot.data)
        self.assertEqual(FAKE['created_at'], sot.created_at)
        self.assertEqual(FAKE['updated_at'], sot.updated_at)


class TestPolicyValidate(base.TestCase):

    def setUp(self):
        super(TestPolicyValidate, self).setUp()

    def test_basic(self):
        sot = policy.PolicyValidate()
        self.assertEqual('policy', sot.resource_key)
        self.assertEqual('policies', sot.resources_key)
        self.assertEqual('/policies/validate', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class AbsoluteLimit(resource.Resource):
    #: Properties
    #: The maximum total amount of backups, in gibibytes (GiB).
    max_total_backup_gigabytes = resource.Body(
        "maxTotalBackupGigabytes", type=int)
    #: The maximum number of backups.
    max_total_backups = resource.Body("maxTotalBackups", type=int)
    #: The maximum number of snapshots.
    max_total_snapshots = resource.Body("maxTotalSnapshots", type=int)
    #: The maximum total amount of volumes, in gibibytes (GiB).
    max_total_volume_gigabytes = resource.Body(
        "maxTotalVolumeGigabytes", type=int)
    #: The maximum number of volumes.
    max_total_volumes = resource.Body("maxTotalVolumes", type=int)
    #: The total number of backups gibibytes (GiB) used.
    total_backup_gigabytes_used = resource.Body(
        "totalBackupGigabytesUsed", type=int)
    #: The total number of backups used.
    total_backups_used = resource.Body("totalBackupsUsed", type=int)
    #: The total number of gibibytes (GiB) used.
    total_gigabytes_used = resource.Body("totalGigabytesUsed", type=int)
    #: The total number of snapshots used.
    total_snapshots_used = resource.Body("totalSnapshotsUsed", type=int)
    #: The total number of volumes used.
    total_volumes_used = resource.Body("totalVolumesUsed", type=int)


class RateLimit(resource.Resource):
    #: Properties
    #: Rate limits next availabe time.
    next_available = resource.Body("next-available")
    #: Integer for rate limits remaining.
    remaining = resource.Body("remaining", type=int)
    #: Unit of measurement for the value parameter.
    unit = resource.Body("unit")
    #: Integer number of requests which can be made.
    value = resource.Body("value", type=int)
    #: An HTTP verb (POST, PUT, etc.).
    verb = resource.Body("verb")


class RateLimits(resource.Resource):
    #: Properties
    #: A list of the specific limits that apply to the ``regex`` and ``uri``.
    limits = resource.Body("limit", type=list, list_type=RateLimit)
    #: A regex representing which routes this rate limit applies to.
    regex = resource.Body("regex")
    #: A URI representing which routes this rate limit applies to.
    uri = resource.Body("uri")


class Limit(resource.Resource):
    resource_key = "limits"
    base_path = "/limits"

    # capabilities
    allow_fetch = True

    #: Properties
    #: An absolute limits object.
    absolute = resource.Body("absolute", type=AbsoluteLimit)
    #: Rate-limit volume copy bandwidth, used to mitigate
    #: slow down of data access from the instances.
    rate = resource.Body("rate", type=list, list_type=RateLimits)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import listener
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'admin_state_up': True,
    'connection_limit': '2',
    'default_pool_id': '3',
    'description': '4',
    'id': IDENTIFIER,
    'loadbalancers': [{'id': '6'}],
    'loadbalancer_id': '6',
    'name': '7',
    'project_id': '8',
    'protocol': '9',
    'protocol_port': '10',
    'default_tls_container_ref': '11',
    'sni_container_refs': [],
}


class TestListener(base.TestCase):

    def test_basic(self):
        sot = listener.Listener()
        self.assertEqual('listener', sot.resource_key)
        self.assertEqual('listeners', sot.resources_key)
        self.assertEqual('/lbaas/listeners', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = listener.Listener(**EXAMPLE)
        self.assertTrue(sot.is_admin_state_up)
        self.assertEqual(EXAMPLE['connection_limit'], sot.connection_limit)
        self.assertEqual(EXAMPLE['default_pool_id'], sot.default_pool_id)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['loadbalancers'], sot.load_balancer_ids)
        self.assertEqual(EXAMPLE['loadbalancer_id'], sot.load_balancer_id)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['protocol'], sot.protocol)
        self.assertEqual(EXAMPLE['protocol_port'], sot.protocol_port)
        self.assertEqual(EXAMPLE['default_tls_container_ref'],
                         sot.default_tls_container_ref)
        self.assertEqual(EXAMPLE['sni_container_refs'],
                         sot.sni_container_refs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.network.v2 import network
from openstack.network.v2 import port
from openstack.network.v2 import trunk as _trunk
from openstack.tests.functional import base


class TestTrunk(base.BaseFunctionalTest):

    TIMEOUT_SCALING_FACTOR = 2.0

    def setUp(self):
        super(TestTrunk, self).setUp()

        # Skip the tests if trunk extension is not enabled.
        if not self.conn.network.find_extension('trunk'):
            self.skipTest('Network trunk extension disabled')

        self.TRUNK_NAME = self.getUniqueString()
        self.TRUNK_NAME_UPDATED = self.getUniqueString()
        net = self.conn.network.create_network()
        assert isinstance(net, network.Network)
        self.NET_ID = net.id
        prt = self.conn.network.create_port(network_id=self.NET_ID)
        assert isinstance(prt, port.Port)
        self.PORT_ID = prt.id
        self.ports_to_clean = [self.PORT_ID]
        trunk = self.conn.network.create_trunk(
            name=self.TRUNK_NAME,
            port_id=self.PORT_ID)
        assert isinstance(trunk, _trunk.Trunk)
        self.TRUNK_ID = trunk.id

    def tearDown(self):
        self.conn.network.delete_trunk(self.TRUNK_ID, ignore_missing=False)
        for port_id in self.ports_to_clean:
            self.conn.network.delete_port(port_id, ignore_missing=False)
        self.conn.network.delete_network(self.NET_ID, ignore_missing=False)
        super(TestTrunk, self).tearDown()

    def test_find(self):
        sot = self.conn.network.find_trunk(self.TRUNK_NAME)
        self.assertEqual(self.TRUNK_ID, sot.id)

    def test_get(self):
        sot = self.conn.network.get_trunk(self.TRUNK_ID)
        self.assertEqual(self.TRUNK_ID, sot.id)
        self.assertEqual(self.TRUNK_NAME, sot.name)

    def test_list(self):
        ids = [o.id for o in self.conn.network.trunks()]
        self.assertIn(self.TRUNK_ID, ids)

    def test_update(self):
        sot = self.conn.network.update_trunk(self.TRUNK_ID,
                                             name=self.TRUNK_NAME_UPDATED)
        self.assertEqual(self.TRUNK_NAME_UPDATED, sot.name)

    def test_subports(self):
        port_for_subport = self.conn.network.create_port(
            network_id=self.NET_ID)
        self.ports_to_clean.append(port_for_subport.id)
        subports = [{
            'port_id': port_for_subport.id,
            'segmentation_type': 'vlan',
            'segmentation_id': 111
        }]

        sot = self.conn.network.get_trunk_subports(self.TRUNK_ID)
        self.assertEqual({'sub_ports': []}, sot)

        self.conn.network.add_trunk_subports(self.TRUNK_ID, subports)
        sot = self.conn.network.get_trunk_subports(self.TRUNK_ID)
        self.assertEqual({'sub_ports': subports}, sot)

        self.conn.network.delete_trunk_subports(
            self.TRUNK_ID, [{'port_id': port_for_subport.id}])
        sot = self.conn.network.get_trunk_subports(self.TRUNK_ID)
        self.assertEqual({'sub_ports': []}, sot)



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import fixtures
from keystoneauth1.fixture import v2
from keystoneauth1.fixture import v3
import os_service_types

_service_type_manager = os_service_types.ServiceTypes()

_SUBURL_TEMPLATES = {
    'public': 'https://example.com/{service_type}',
    'internal': 'https://internal.example.com/{service_type}',
    'admin': 'https://example.com/{service_type}',
}
_ENDPOINT_TEMPLATES = {
    'public': 'https://{service_type}.example.com',
    'internal': 'https://internal.{service_type}.example.com',
    'admin': 'https://{service_type}.example.com',
}


class ConnectionFixture(fixtures.Fixture):

    _suffixes = {
        'baremetal': '/',
        'block-storage': '/{project_id}',
        'compute': '/v2.1/',
        'container-infrastructure-management': '/v1',
        'object-store': '/v1/{project_id}',
        'orchestration': '/v1/{project_id}',
        'volumev2': '/v2/{project_id}',
        'volumev3': '/v3/{project_id}',
    }

    def __init__(self, suburl=False, project_id=None, *args, **kwargs):
        super(ConnectionFixture, self).__init__(*args, **kwargs)
        self._endpoint_templates = _ENDPOINT_TEMPLATES
        if suburl:
            self.use_suburl()
        self.project_id = project_id or uuid.uuid4().hex.replace('-', '')
        self.build_tokens()

    def use_suburl(self):
        self._endpoint_templates = _SUBURL_TEMPLATES

    def _get_endpoint_templates(self, service_type, alias=None, v2=False):
        templates = {}
        for k, v in self._endpoint_templates.items():
            suffix = self._suffixes.get(
                alias, self._suffixes.get(service_type, ''))
            # For a keystone v2 catalog, we want to list the
            # versioned endpoint in the catalog, because that's
            # more likely how those were deployed.
            if v2:
                suffix = '/v2.0'
            templates[k] = (v + suffix).format(
                service_type=service_type,
                project_id=self.project_id,
            )
        return templates

    def _setUp(self):
        pass

    def clear_tokens(self):
        self.v2_token = v2.Token(tenant_id=self.project_id)
        self.v3_token = v3.Token(project_id=self.project_id)

    def build_tokens(self):
        self.clear_tokens()
        for service in _service_type_manager.services:
            service_type = service['service_type']
            if service_type == 'ec2-api':
                continue
            service_name = service['project']
            ets = self._get_endpoint_templates(service_type)
            v3_svc = self.v3_token.add_service(
                service_type, name=service_name)
            v2_svc = self.v2_token.add_service(
                service_type, name=service_name)
            v3_svc.add_standard_endpoints(region='RegionOne', **ets)
            if service_type == 'identity':
                ets = self._get_endpoint_templates(service_type, v2=True)
            v2_svc.add_endpoint(region='RegionOne', **ets)
            for alias in service.get('aliases', []):
                ets = self._get_endpoint_templates(service_type, alias=alias)
                v3_svc = self.v3_token.add_service(alias, name=service_name)
                v2_svc = self.v2_token.add_service(alias, name=service_name)
                v3_svc.add_standard_endpoints(region='RegionOne', **ets)
                v2_svc.add_endpoint(region='RegionOne', **ets)

    def _cleanup(self):
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import rbac_policy
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'action': 'access_as_shared',
    'object_id': IDENTIFIER,
    'object_type': 'network',
    'target_tenant': '10',
    'project_id': '5',
}


class TestRBACPolicy(base.TestCase):

    def test_basic(self):
        sot = rbac_policy.RBACPolicy()
        self.assertEqual('rbac_policy', sot.resource_key)
        self.assertEqual('rbac_policies', sot.resources_key)
        self.assertEqual('/rbac-policies', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                'action': 'action',
                'object_id': 'object_id',
                'object_type': 'object_type',
                'project_id': 'project_id',
                'target_project_id': 'target_tenant',
                'limit': 'limit',
                'marker': 'marker',
            },
            sot._query_mapping._mapping)

    def test_make_it(self):
        sot = rbac_policy.RBACPolicy(**EXAMPLE)
        self.assertEqual(EXAMPLE['action'], sot.action)
        self.assertEqual(EXAMPLE['object_id'], sot.object_id)
        self.assertEqual(EXAMPLE['object_type'], sot.object_type)
        self.assertEqual(EXAMPLE['target_tenant'], sot.target_project_id)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from testtools import matchers

from openstack.cloud import exc
from openstack.tests.unit import base


class TestRoleAssignment(base.TestCase):

    def _build_role_assignment_response(self, role_id, scope_type, scope_id,
                                        entity_type, entity_id):
        self.assertThat(['group', 'user'], matchers.Contains(entity_type))
        self.assertThat(['project', 'domain'], matchers.Contains(scope_type))
        # NOTE(notmorgan): Links are thrown out by shade, but we construct them
        # for corectness.
        link_str = ('https://identity.example.com/identity/v3/{scope_t}s'
                    '/{scopeid}/{entity_t}s/{entityid}/roles/{roleid}')
        return [{
            'links': {'assignment': link_str.format(
                scope_t=scope_type, scopeid=scope_id, entity_t=entity_type,
                entityid=entity_id, roleid=role_id)},
            'role': {'id': role_id},
            'scope': {scope_type: {'id': scope_id}},
            entity_type: {'id': entity_id}
        }]

    def setUp(self, cloud_config_fixture='clouds.yaml'):
        super(TestRoleAssignment, self).setUp(cloud_config_fixture)
        self.role_data = self._get_role_data()
        self.domain_data = self._get_domain_data()
        self.user_data = self._get_user_data(
            domain_id=self.domain_data.domain_id)
        self.project_data = self._get_project_data(
            domain_id=self.domain_data.domain_id)
        self.project_data_v2 = self._get_project_data(
            project_name=self.project_data.project_name,
            project_id=self.project_data.project_id,
            v3=False)
        self.group_data = self._get_group_data(
            domain_id=self.domain_data.domain_id)

        self.user_project_assignment = self._build_role_assignment_response(
            role_id=self.role_data.role_id, scope_type='project',
            scope_id=self.project_data.project_id, entity_type='user',
            entity_id=self.user_data.user_id)

        self.group_project_assignment = self._build_role_assignment_response(
            role_id=self.role_data.role_id, scope_type='project',
            scope_id=self.project_data.project_id, entity_type='group',
            entity_id=self.group_data.group_id)

        self.user_domain_assignment = self._build_role_assignment_response(
            role_id=self.role_data.role_id, scope_type='domain',
            scope_id=self.domain_data.domain_id, entity_type='user',
            entity_id=self.user_data.user_id)

        self.group_domain_assignment = self._build_role_assignment_response(
            role_id=self.role_data.role_id, scope_type='domain',
            scope_id=self.domain_data.domain_id, entity_type='group',
            entity_id=self.group_data.group_id)

        # Cleanup of instances to ensure garbage collection/no leaking memory
        # in tests.
        self.addCleanup(delattr, self, 'role_data')
        self.addCleanup(delattr, self, 'user_data')
        self.addCleanup(delattr, self, 'domain_data')
        self.addCleanup(delattr, self, 'group_data')
        self.addCleanup(delattr, self, 'project_data')
        self.addCleanup(delattr, self, 'project_data_v2')
        self.addCleanup(delattr, self, 'user_project_assignment')
        self.addCleanup(delattr, self, 'group_project_assignment')
        self.addCleanup(delattr, self, 'user_domain_assignment')
        self.addCleanup(delattr, self, 'group_domain_assignment')

    def get_mock_url(self, service_type='identity', interface='public',
                     resource='role_assignments', append=None,
                     base_url_append='v3', qs_elements=None):
        return super(TestRoleAssignment, self).get_mock_url(
            service_type, interface, resource, append, base_url_append,
            qs_elements)

    def __get(self, resource, data, attr, qs_elements, use_name=False,
              is_found=True):
        if not use_name:
            if is_found:
                return [dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource=resource + 's',  # do roles from role
                        append=[getattr(data, attr)],
                        qs_elements=qs_elements),
                    status_code=200,
                    json=data.json_response)
                ]
            else:
                return [
                    dict(
                        method='GET',
                        uri=self.get_mock_url(
                            resource=resource + 's',  # do roles from role
                            append=[getattr(data, attr)],
                            qs_elements=qs_elements),
                        status_code=404),
                    dict(
                        method='GET',
                        uri=self.get_mock_url(
                            resource=resource + 's',  # do roles from role
                            qs_elements=qs_elements),
                        status_code=200,
                        json={(resource + 's'): []})
                ]
        else:
            return [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource=resource + 's',
                        append=[getattr(data, attr)],
                        qs_elements=qs_elements),
                    status_code=404),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        resource=resource + 's',
                        qs_elements=[
                            'name=' + getattr(data, attr)] + qs_elements),
                    status_code=200,
                    json={
                        (resource + 's'): [data.json_response[resource]]})
            ]

    def __user_mocks(self, user_data, use_name, is_found=True):
        uri_mocks = []
        if not use_name:
            uri_mocks.append(dict(
                method='GET',
                uri=self.get_mock_url(resource='users'),
                status_code=200,
                json={'users': [user_data.json_response['user']] if is_found
                      else []}))
        else:
            uri_mocks.append(dict(
                method='GET',
                uri=self.get_mock_url(
                    resource='users',
                    qs_elements=[
                        'name=' + user_data.name
                    ]
                ),
                status_code=200,
                json={'users': [user_data.json_response['user']] if is_found
                      else []}))
        return uri_mocks

    def _get_mock_role_query_urls(
        self, role_data, domain_data=None, project_data=None,
        group_data=None, user_data=None,
        use_role_name=False, use_domain_name=False, use_project_name=False,
        use_group_name=False, use_user_name=False, use_domain_in_query=True
    ):
        """Build uri mocks for querying role assignments
        """
        uri_mocks = []

        if domain_data:
            uri_mocks.extend(
                self.__get(
                    'domain', domain_data,
                    'domain_id' if not use_domain_name else 'domain_name',
                    [], use_name=use_domain_name)
            )

        qs_elements = []
        if domain_data and use_domain_in_query:
            qs_elements = [
                'domain_id=' + domain_data.domain_id
            ]

        uri_mocks.extend(
            self.__get(
                'role', role_data,
                'role_id' if not use_role_name else 'role_name',
                [], use_name=use_role_name)
        )

        if user_data:
            uri_mocks.extend(self.__user_mocks(
                user_data, use_user_name, is_found=True))

        if group_data:
            uri_mocks.extend(
                self.__get(
                    'group', group_data,
                    'group_id' if not use_group_name else 'group_name',
                    qs_elements, use_name=use_group_name)
            )

        if project_data:
            uri_mocks.extend(
                self.__get(
                    'project', project_data,
                    'project_id' if not use_project_name else 'project_name',
                    qs_elements, use_name=use_project_name)
            )

        return uri_mocks

    def test_grant_role_user_id_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data, project_data=self.project_data,
            user_data=self.user_data, use_role_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.user_id,
                project=self.project_data.project_id))
        self.assert_calls()

    def test_grant_role_user_name_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, user_data=self.user_data,
            use_role_name=True, use_user_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_id))

    def test_grant_role_user_id_project_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, user_data=self.user_data)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.grant_role(
            self.role_data.role_id,
            user=self.user_data.user_id,
            project=self.project_data.project_id))
        self.assert_calls()

    def test_grant_role_user_name_project_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, user_data=self.user_data,
            use_role_name=True, use_user_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.grant_role(
            self.role_data.role_name,
            user=self.user_data.name,
            project=self.project_data.project_id))
        self.assert_calls()

    def test_grant_role_group_id_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, group_data=self.group_data,
            use_role_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'groups', self.group_data.group_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_id,
                project=self.project_data.project_id))
        self.assert_calls()

    def test_grant_role_group_name_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, group_data=self.group_data,
            use_role_name=True, use_group_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'groups', self.group_data.group_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                project=self.project_data.project_id))
        self.assert_calls()

    def test_grant_role_group_id_project_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, group_data=self.group_data)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.grant_role(
            self.role_data.role_id,
            group=self.group_data.group_id,
            project=self.project_data.project_id))
        self.assert_calls()

    def test_grant_role_group_name_project_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, group_data=self.group_data,
            use_role_name=True, use_group_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.grant_role(
            self.role_data.role_name,
            group=self.group_data.group_name,
            project=self.project_data.project_id))
        self.assert_calls()

# ===== Domain
    def test_grant_role_user_id_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data, domain_data=self.domain_data,
            user_data=self.user_data, use_role_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[self.domain_data.domain_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.user_id,
                domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_grant_role_user_name_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, user_data=self.user_data,
            use_role_name=True, use_user_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[self.domain_data.domain_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain=self.domain_data.domain_id))

    def test_grant_role_user_id_domain_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, user_data=self.user_data)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.grant_role(
            self.role_data.role_id,
            user=self.user_data.user_id,
            domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_grant_role_user_name_domain_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, user_data=self.user_data,
            use_role_name=True, use_user_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.grant_role(
            self.role_data.role_name,
            user=self.user_data.name,
            domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_grant_role_group_id_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, group_data=self.group_data,
            use_role_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[self.domain_data.domain_id,
                             'groups', self.group_data.group_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_id,
                domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_grant_role_group_name_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, group_data=self.group_data,
            use_role_name=True, use_group_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[self.domain_data.domain_id,
                             'groups', self.group_data.group_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_grant_role_group_id_domain_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, group_data=self.group_data)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.grant_role(
            self.role_data.role_id,
            group=self.group_data.group_id,
            domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_grant_role_group_name_domain_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, group_data=self.group_data,
            use_role_name=True, use_group_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.grant_role(
            self.role_data.role_name,
            group=self.group_data.group_name,
            domain=self.domain_data.domain_id))
        self.assert_calls()

# ==== Revoke
    def test_revoke_role_user_id_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data, project_data=self.project_data,
            user_data=self.user_data, use_role_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.user_id,
                project=self.project_data.project_id))
        self.assert_calls()

    def test_revoke_role_user_name_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, user_data=self.user_data,
            use_role_name=True, use_user_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_id))

    def test_revoke_role_user_id_project_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, user_data=self.user_data)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.revoke_role(
            self.role_data.role_id,
            user=self.user_data.user_id,
            project=self.project_data.project_id))
        self.assert_calls()

    def test_revoke_role_user_name_project_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, user_data=self.user_data,
            use_role_name=True, use_user_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.revoke_role(
            self.role_data.role_name,
            user=self.user_data.name,
            project=self.project_data.project_id))
        self.assert_calls()

    def test_revoke_role_group_id_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, group_data=self.group_data,
            use_role_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'groups', self.group_data.group_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_id,
                project=self.project_data.project_id))
        self.assert_calls()

    def test_revoke_role_group_name_project(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, group_data=self.group_data,
            use_role_name=True, use_group_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'groups', self.group_data.group_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                project=self.project_data.project_id))
        self.assert_calls()

    def test_revoke_role_group_id_project_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, group_data=self.group_data)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.revoke_role(
            self.role_data.role_id,
            group=self.group_data.group_id,
            project=self.project_data.project_id))
        self.assert_calls()

    def test_revoke_role_group_name_project_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, group_data=self.group_data,
            use_role_name=True, use_group_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.revoke_role(
            self.role_data.role_name,
            group=self.group_data.group_name,
            project=self.project_data.project_id))
        self.assert_calls()

# ==== Domain
    def test_revoke_role_user_id_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data, domain_data=self.domain_data,
            user_data=self.user_data, use_role_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[self.domain_data.domain_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.user_id,
                domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_revoke_role_user_name_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, user_data=self.user_data,
            use_role_name=True, use_user_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[self.domain_data.domain_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain=self.domain_data.domain_id))

    def test_revoke_role_user_id_domain_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, user_data=self.user_data)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.revoke_role(
            self.role_data.role_id,
            user=self.user_data.user_id,
            domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_revoke_role_user_name_domain_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, user_data=self.user_data,
            use_role_name=True, use_user_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.revoke_role(
            self.role_data.role_name,
            user=self.user_data.name,
            domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_revoke_role_group_id_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, group_data=self.group_data,
            use_role_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[self.domain_data.domain_id,
                             'groups', self.group_data.group_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_id,
                domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_revoke_role_group_name_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, group_data=self.group_data,
            use_role_name=True, use_group_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[self.domain_data.domain_id,
                             'groups', self.group_data.group_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_revoke_role_group_id_domain_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, group_data=self.group_data)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.revoke_role(
            self.role_data.role_id,
            group=self.group_data.group_id,
            domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_revoke_role_group_name_domain_not_exists(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            domain_data=self.domain_data, group_data=self.group_data,
            use_role_name=True, use_group_name=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='domains',
                     append=[
                         self.domain_data.domain_id,
                         'groups', self.group_data.group_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
        ])
        self.register_uris(uris)

        self.assertFalse(self.cloud.revoke_role(
            self.role_data.role_name,
            group=self.group_data.group_name,
            domain=self.domain_data.domain_id))
        self.assert_calls()

    def test_grant_no_role(self):
        uris = self.__get(
            'domain', self.domain_data, 'domain_name', [], use_name=True)
        uris.extend([
            dict(method='GET',
                 uri=self.get_mock_url(
                     resource='roles',
                     append=[self.role_data.role_name],
                 ),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     resource='roles',
                     qs_elements=[
                         'name=' + self.role_data.role_name,
                     ]),
                 status_code=200,
                 json={'roles': []})
        ])
        self.register_uris(uris)

        with testtools.ExpectedException(
            exc.OpenStackCloudException,
            'Role {0} not found'.format(self.role_data.role_name)
        ):
            self.cloud.grant_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_name)
        self.assert_calls()

    def test_revoke_no_role(self):
        uris = self.__get(
            'domain', self.domain_data, 'domain_name', [], use_name=True)
        uris.extend([
            dict(method='GET',
                 uri=self.get_mock_url(
                     resource='roles',
                     append=[self.role_data.role_name],
                 ),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     resource='roles',
                     qs_elements=[
                         'name=' + self.role_data.role_name,
                     ]),
                 status_code=200,
                 json={'roles': []})
        ])
        self.register_uris(uris)

        with testtools.ExpectedException(
            exc.OpenStackCloudException,
            'Role {0} not found'.format(self.role_data.role_name)
        ):
            self.cloud.revoke_role(
                self.role_data.role_name,
                group=self.group_data.group_name,
                domain=self.domain_data.domain_name)
        self.assert_calls()

    def test_grant_no_user_or_group_specified(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True)
        self.register_uris(uris)
        with testtools.ExpectedException(
                exc.OpenStackCloudException,
                'Must specify either a user or a group'
        ):
            self.cloud.grant_role(self.role_data.role_name)
        self.assert_calls()

    def test_revoke_no_user_or_group_specified(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True)
        self.register_uris(uris)
        with testtools.ExpectedException(
                exc.OpenStackCloudException,
                'Must specify either a user or a group'
        ):
            self.cloud.revoke_role(self.role_data.role_name)
        self.assert_calls()

    def test_grant_no_user_or_group(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True)
        uris.extend(self.__user_mocks(self.user_data, use_name=True,
                                      is_found=False))
        self.register_uris(uris)

        with testtools.ExpectedException(
            exc.OpenStackCloudException,
            'Must specify either a user or a group'
        ):
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name)
        self.assert_calls()

    def test_revoke_no_user_or_group(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True)
        uris.extend(self.__user_mocks(self.user_data, use_name=True,
                                      is_found=False))
        self.register_uris(uris)

        with testtools.ExpectedException(
            exc.OpenStackCloudException,
            'Must specify either a user or a group'
        ):
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name)
        self.assert_calls()

    def test_grant_both_user_and_group(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True)
        uris.extend(self.__user_mocks(self.user_data, use_name=True))
        uris.extend(self.__get(
            'group', self.group_data, 'group_name', [], use_name=True))
        self.register_uris(uris)

        with testtools.ExpectedException(
            exc.OpenStackCloudException,
            'Specify either a group or a user, not both'
        ):
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                group=self.group_data.group_name)
        self.assert_calls()

    def test_revoke_both_user_and_group(self):
        uris = self.__get(
            'role', self.role_data, 'role_name', [], use_name=True)
        uris.extend(self.__user_mocks(self.user_data, use_name=True))
        uris.extend(self.__get(
            'group', self.group_data, 'group_name', [], use_name=True))
        self.register_uris(uris)

        with testtools.ExpectedException(
            exc.OpenStackCloudException,
            'Specify either a group or a user, not both'
        ):
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                group=self.group_data.group_name)

    def test_grant_both_project_and_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, user_data=self.user_data,
            domain_data=self.domain_data,
            use_role_name=True, use_user_name=True, use_project_name=True,
            use_domain_name=True, use_domain_in_query=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=404),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_name,
                domain=self.domain_data.domain_name))

    def test_revoke_both_project_and_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            project_data=self.project_data, user_data=self.user_data,
            domain_data=self.domain_data,
            use_role_name=True, use_user_name=True, use_project_name=True,
            use_domain_name=True, use_domain_in_query=True)
        uris.extend([
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[
                         self.project_data.project_id,
                         'users', self.user_data.user_id,
                         'roles', self.role_data.role_id
                     ]),
                 complete_qs=True,
                 status_code=204),
            dict(method='DELETE',
                 uri=self.get_mock_url(
                     resource='projects',
                     append=[self.project_data.project_id,
                             'users', self.user_data.user_id,
                             'roles', self.role_data.role_id]),
                 status_code=200),
        ])
        self.register_uris(uris)

        self.assertTrue(
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                project=self.project_data.project_name,
                domain=self.domain_data.domain_name))

    def test_grant_no_project_or_domain(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            user_data=self.user_data,
            use_role_name=True, use_user_name=True)

        self.register_uris(uris)

        with testtools.ExpectedException(
            exc.OpenStackCloudException,
            'Must specify either a domain, project or system'
        ):
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name)
        self.assert_calls()

    def test_revoke_no_project_or_domain_or_system(self):
        uris = self._get_mock_role_query_urls(
            self.role_data,
            user_data=self.user_data,
            use_role_name=True, use_user_name=True)

        self.register_uris(uris)

        with testtools.ExpectedException(
            exc.OpenStackCloudException,
            'Must specify either a domain, project or system'
        ):
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name)
        self.assert_calls()

    def test_grant_bad_domain_exception(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     resource='domains', append=['baddomain']),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     resource='domains',
                     qs_elements=['name=baddomain']),
                 status_code=404)
        ])
        with testtools.ExpectedException(exc.OpenStackCloudURINotFound):
            self.cloud.grant_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain='baddomain')
        self.assert_calls()

    def test_revoke_bad_domain_exception(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     resource='domains', append=['baddomain']),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     resource='domains',
                     qs_elements=['name=baddomain']),
                 status_code=404)
        ])
        with testtools.ExpectedException(exc.OpenStackCloudURINotFound):
            self.cloud.revoke_role(
                self.role_data.role_name,
                user=self.user_data.name,
                domain='baddomain')
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.unit import base


class TestLimits(base.TestCase):

    def test_get_compute_limits(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'compute', 'public', append=['limits']),
                 json={
                     "limits": {
                         "absolute": {
                             "maxImageMeta": 128,
                             "maxPersonality": 5,
                             "maxPersonalitySize": 10240,
                             "maxSecurityGroupRules": 20,
                             "maxSecurityGroups": 10,
                             "maxServerMeta": 128,
                             "maxTotalCores": 20,
                             "maxTotalFloatingIps": 10,
                             "maxTotalInstances": 10,
                             "maxTotalKeypairs": 100,
                             "maxTotalRAMSize": 51200,
                             "maxServerGroups": 10,
                             "maxServerGroupMembers": 10,
                             "totalCoresUsed": 0,
                             "totalInstancesUsed": 0,
                             "totalRAMUsed": 0,
                             "totalSecurityGroupsUsed": 0,
                             "totalFloatingIpsUsed": 0,
                             "totalServerGroupsUsed": 0
                         },
                         "rate": []
                     }
                 }),
        ])

        self.cloud.get_compute_limits()

        self.assert_calls()

    def test_other_get_compute_limits(self):
        project = self.mock_for_keystone_projects(project_count=1,
                                                  list_get=True)[0]
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'compute', 'public', append=['limits'],
                     qs_elements=[
                         'tenant_id={id}'.format(id=project.project_id)
                     ]),
                 json={
                     "limits": {
                         "absolute": {
                             "maxImageMeta": 128,
                             "maxPersonality": 5,
                             "maxPersonalitySize": 10240,
                             "maxSecurityGroupRules": 20,
                             "maxSecurityGroups": 10,
                             "maxServerMeta": 128,
                             "maxTotalCores": 20,
                             "maxTotalFloatingIps": 10,
                             "maxTotalInstances": 10,
                             "maxTotalKeypairs": 100,
                             "maxTotalRAMSize": 51200,
                             "maxServerGroups": 10,
                             "maxServerGroupMembers": 10,
                             "totalCoresUsed": 0,
                             "totalInstancesUsed": 0,
                             "totalRAMUsed": 0,
                             "totalSecurityGroupsUsed": 0,
                             "totalFloatingIpsUsed": 0,
                             "totalServerGroupsUsed": 0
                         },
                         "rate": []
                     }
                 }),
        ])

        self.cloud.get_compute_limits(project.project_id)

        self.assert_calls()



# Copyright(c) 2018 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack.instance_ha.v1 import _proxy
from openstack.instance_ha.v1 import host
from openstack.instance_ha.v1 import notification
from openstack.instance_ha.v1 import segment
from openstack.tests.unit import test_proxy_base

SEGMENT_ID = "c50b96eb-2a66-40f8-bca8-c5fa90d595c0"
HOST_ID = "52d05e43-d08e-42b8-ae33-e47c8ea2ad47"


class TestInstanceHaProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super(TestInstanceHaProxy, self).setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestInstanceHaHosts(TestInstanceHaProxy):
    def test_hosts(self):
        self.verify_list(self.proxy.hosts,
                         host.Host,
                         method_args=[SEGMENT_ID],
                         expected_args=[],
                         expected_kwargs={"segment_id": SEGMENT_ID})

    def test_host_get(self):
        self.verify_get(self.proxy.get_host,
                        host.Host,
                        method_args=[HOST_ID],
                        method_kwargs={"segment_id": SEGMENT_ID},
                        expected_kwargs={"segment_id": SEGMENT_ID})

    def test_host_create(self):
        self.verify_create(self.proxy.create_host,
                           host.Host,
                           method_args=[SEGMENT_ID],
                           method_kwargs={},
                           expected_args=[],
                           expected_kwargs={"segment_id": SEGMENT_ID})

    def test_host_update(self):
        self.verify_update(self.proxy.update_host,
                           host.Host,
                           method_kwargs={"segment_id": SEGMENT_ID})

    def test_host_delete(self):
        self.verify_delete(self.proxy.delete_host,
                           host.Host,
                           True,
                           method_kwargs={"segment_id": SEGMENT_ID},
                           expected_kwargs={"segment_id": SEGMENT_ID})


class TestInstanceHaNotifications(TestInstanceHaProxy):
    def test_notifications(self):
        self.verify_list(self.proxy.notifications, notification.Notification)

    def test_notification_get(self):
        self.verify_get(self.proxy.get_notification,
                        notification.Notification)

    def test_notification_create(self):
        self.verify_create(self.proxy.create_notification,
                           notification.Notification)


class TestInstanceHaSegments(TestInstanceHaProxy):
    def test_segments(self):
        self.verify_list(self.proxy.segments, segment.Segment)

    def test_segment_get(self):
        self.verify_get(self.proxy.get_segment, segment.Segment)

    def test_segment_create(self):
        self.verify_create(self.proxy.create_segment, segment.Segment)

    def test_segment_update(self):
        self.verify_update(self.proxy.update_segment, segment.Segment)

    def test_segment_delete(self):
        self.verify_delete(self.proxy.delete_segment, segment.Segment, True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack.common import metadata
from openstack import resource


class Image(resource.Resource, metadata.MetadataMixin):
    resource_key = 'image'
    resources_key = 'images'
    base_path = '/images'

    # capabilities
    allow_fetch = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        "server", "name", "status", "type",
        min_disk="minDisk",
        min_ram="minRam",
        changes_since="changes-since")

    # Properties
    #: Links pertaining to this image. This is a list of dictionaries,
    #: each including keys ``href`` and ``rel``, and optionally ``type``.
    links = resource.Body('links')
    #: The name of this image.
    name = resource.Body('name')
    #: Timestamp when the image was created.
    created_at = resource.Body('created')
    #: The mimimum disk size. *Type: int*
    min_disk = resource.Body('minDisk', type=int)
    #: The minimum RAM size. *Type: int*
    min_ram = resource.Body('minRam', type=int)
    #: If this image is still building, its progress is represented here.
    #: Once an image is created, progres will be 100. *Type: int*
    progress = resource.Body('progress', type=int)
    #: The status of this image.
    status = resource.Body('status')
    #: Timestamp when the image was updated.
    updated_at = resource.Body('updated')
    #: Size of the image in bytes. *Type: int*
    size = resource.Body('OS-EXT-IMG-SIZE:size', type=int)


ImageDetail = Image



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import queue
import string
import threading
import time

import keystoneauth1
from keystoneauth1 import discover

from openstack import _log
from openstack import exceptions


def urljoin(*args):
    """A custom version of urljoin that simply joins strings into a path.

    The real urljoin takes into account web semantics like when joining a url
    like /path this should be joined to http://host/path as it is an anchored
    link. We generally won't care about that in client.
    """
    return '/'.join(str(a or '').strip('/') for a in args)


def iterate_timeout(timeout, message, wait=2):
    """Iterate and raise an exception on timeout.

    This is a generator that will continually yield and sleep for
    wait seconds, and if the timeout is reached, will raise an exception
    with <message>.

    """
    log = _log.setup_logging('openstack.iterate_timeout')

    try:
        # None as a wait winds up flowing well in the per-resource cache
        # flow. We could spread this logic around to all of the calling
        # points, but just having this treat None as "I don't have a value"
        # seems friendlier
        if wait is None:
            wait = 2
        elif wait == 0:
            # wait should be < timeout, unless timeout is None
            wait = 0.1 if timeout is None else min(0.1, timeout)
        wait = float(wait)
    except ValueError:
        raise exceptions.SDKException(
            "Wait value must be an int or float value. {wait} given"
            " instead".format(wait=wait))

    start = time.time()
    count = 0
    while (timeout is None) or (time.time() < start + timeout):
        count += 1
        yield count
        log.debug('Waiting %s seconds', wait)
        time.sleep(wait)
    raise exceptions.ResourceTimeout(message)


def get_string_format_keys(fmt_string, old_style=True):
    """Gets a list of required keys from a format string

    Required mostly for parsing base_path urls for required keys, which
    use the old style string formatting.
    """
    if old_style:
        class AccessSaver:
            def __init__(self):
                self.keys = []

            def __getitem__(self, key):
                self.keys.append(key)

        a = AccessSaver()
        fmt_string % a

        return a.keys
    else:
        keys = []
        for t in string.Formatter().parse(fmt_string):
            if t[1] is not None:
                keys.append(t[1])
        return keys


def supports_microversion(adapter, microversion, raise_exception=False):
    """Determine if the given adapter supports the given microversion.

    Checks the min and max microversion asserted by the service and checks to
    make sure that ``min <= microversion <= max``. Current default microversion
    is taken into consideration if set and verifies that ``microversion <=
    default``.

    :param adapter: :class:`~keystoneauth1.adapter.Adapter` instance.
    :param str microversion: String containing the desired microversion.
    :param bool raise_exception: Raise exception when requested microversion
        is not supported be the server side or is higher than the current
        default microversion.
    :returns: True if the service supports the microversion.
    :rtype: bool
    :raises: :class:`~openstack.exceptions.SDKException` when requested
        microversion is not supported.
    """

    endpoint_data = adapter.get_endpoint_data()
    if (endpoint_data.min_microversion
            and endpoint_data.max_microversion
            and discover.version_between(
                endpoint_data.min_microversion,
                endpoint_data.max_microversion,
                microversion)):
        if adapter.default_microversion is not None:
            # If default_microversion is set - evaluate
            # whether it match the expectation
            candidate = discover.normalize_version_number(
                adapter.default_microversion)
            required = discover.normalize_version_number(microversion)
            supports = discover.version_match(required, candidate)
            if raise_exception and not supports:
                raise exceptions.SDKException(
                    'Required microversion {ver} is higher than currently '
                    'selected {curr}'.format(
                        ver=microversion,
                        curr=adapter.default_microversion)
                )
            return supports
        return True
    if raise_exception:
        raise exceptions.SDKException(
            'Required microversion {ver} is not supported '
            'by the server side'.format(ver=microversion)
        )
    return False


def require_microversion(adapter, required):
    """Require microversion.

    :param adapter: :class:`~keystoneauth1.adapter.Adapter` instance.
    :param str microversion: String containing the desired microversion.
    :raises: :class:`~openstack.exceptions.SDKException` when requested
        microversion is not supported
    """
    supports_microversion(adapter, required, raise_exception=True)


def pick_microversion(session, required):
    """Get a new microversion if it is higher than session's default.

    :param session: The session to use for making this request.
    :type session: :class:`~keystoneauth1.adapter.Adapter`
    :param required: Minimum version that is required for an action.
    :type required: String or tuple or None.
    :return: ``required`` as a string if the ``session``'s default is too low,
        otherwise the ``session``'s default. Returns ``None`` if both
        are ``None``.
    :raises: TypeError if ``required`` is invalid.
    :raises: :class:`~openstack.exceptions.SDKException` if requested
        microversion is not supported.
    """
    if required is not None:
        required = discover.normalize_version_number(required)

    if session.default_microversion is not None:
        default = discover.normalize_version_number(
            session.default_microversion)

        if required is None:
            required = default
        else:
            required = (default if discover.version_match(required, default)
                        else required)

    if required is not None:
        if not supports_microversion(session, required):
            raise exceptions.SDKException(
                'Requested microversion is not supported by the server side '
                'or the default microversion is too low')
        return discover.version_to_string(required)


def maximum_supported_microversion(adapter, client_maximum):
    """Determine the maximum microversion supported by both client and server.

    :param adapter: :class:`~keystoneauth1.adapter.Adapter` instance.
    :param client_maximum: Maximum microversion supported by the client.
        If ``None``, ``None`` is returned.

    :returns: the maximum supported microversion as string or ``None``.
    """
    if client_maximum is None:
        return None

    # NOTE(dtantsur): if we cannot determine supported microversions, fall back
    # to the default one.
    try:
        endpoint_data = adapter.get_endpoint_data()
    except keystoneauth1.exceptions.discovery.DiscoveryFailure:
        endpoint_data = None

    if endpoint_data is None:
        log = _log.setup_logging('openstack')
        log.warning('Cannot determine endpoint data for service %s',
                    adapter.service_type or adapter.service_name)
        return None

    if not endpoint_data.max_microversion:
        return None

    client_max = discover.normalize_version_number(client_maximum)
    server_max = discover.normalize_version_number(
        endpoint_data.max_microversion)

    if endpoint_data.min_microversion:
        server_min = discover.normalize_version_number(
            endpoint_data.min_microversion)
        if client_max < server_min:
            # NOTE(dtantsur): we may want to raise in this case, but this keeps
            # the current behavior intact.
            return None

    result = min(client_max, server_max)
    return discover.version_to_string(result)


def _hashes_up_to_date(md5, sha256, md5_key, sha256_key):
    '''Compare md5 and sha256 hashes for being up to date

    md5 and sha256 are the current values.
    md5_key and sha256_key are the previous values.
    '''
    up_to_date = False
    if md5 and md5_key == md5:
        up_to_date = True
    if sha256 and sha256_key == sha256:
        up_to_date = True
    if md5 and md5_key != md5:
        up_to_date = False
    if sha256 and sha256_key != sha256:
        up_to_date = False
    return up_to_date


try:
    _test_md5 = hashlib.md5(usedforsecurity=False)  # nosec

    # Python distributions that support a hashlib.md5 with the usedforsecurity
    # keyword can just use that md5 definition as-is
    # See https://bugs.python.org/issue9216
    #
    # TODO(alee) Remove this wrapper when the minimum python version is bumped
    # to 3.9 (which is the first upstream version to support this keyword)
    # See https://docs.python.org/3.9/library/hashlib.html
    md5 = hashlib.md5
except TypeError:
    def md5(string=b'', usedforsecurity=True):
        """Return an md5 hashlib object without usedforsecurity parameter
        For python distributions that do not yet support this keyword
        parameter, we drop the parameter
        """
        return hashlib.md5(string)  # nosec


def _calculate_data_hashes(data):
    _md5 = md5(usedforsecurity=False)
    _sha256 = hashlib.sha256()

    if hasattr(data, 'read'):
        for chunk in iter(lambda: data.read(8192), b''):
            _md5.update(chunk)
            _sha256.update(chunk)
    else:
        _md5.update(data)
        _sha256.update(data)
    return (_md5.hexdigest(), _sha256.hexdigest())


def _get_file_hashes(filename):
    (_md5, _sha256) = (None, None)
    with open(filename, 'rb') as file_obj:
        (_md5, _sha256) = _calculate_data_hashes(file_obj)

    return (_md5, _sha256)


class TinyDAG:
    """Tiny DAG

    Bases on the Kahn's algorithm, and enables parallel visiting of the nodes
    (parallel execution of the workflow items).
    """

    def __init__(self, data=None):
        self._reset()
        self._lock = threading.Lock()
        if data and isinstance(data, dict):
            self.from_dict(data)

    def _reset(self):
        self._graph = dict()
        self._wait_timeout = 120

    @property
    def graph(self):
        """Get graph as adjacency dict
        """
        return self._graph

    def add_node(self, node):
        self._graph.setdefault(node, set())

    def add_edge(self, u, v):
        self._graph[u].add(v)

    def from_dict(self, data):
        self._reset()
        for k, v in data.items():
            self.add_node(k)
            for dep in v:
                self.add_edge(k, dep)

    def walk(self, timeout=None):
        """Start the walking from the beginning.
        """
        if timeout:
            self._wait_timeout = timeout
        return self

    def __iter__(self):
        self._start_traverse()
        return self

    def __next__(self):
        # Start waiting if it is expected to get something
        # (counting down from graph length to 0).
        if (self._it_cnt > 0):
            self._it_cnt -= 1
            try:
                res = self._queue.get(
                    block=True,
                    timeout=self._wait_timeout)
                return res

            except queue.Empty:
                raise exceptions.SDKException('Timeout waiting for '
                                              'cleanup task to complete')
        else:
            raise StopIteration

    def node_done(self, node):
        """Mark node as "processed" and put following items into the queue"""
        self._done.add(node)

        for v in self._graph[node]:
            self._run_in_degree[v] -= 1
            if self._run_in_degree[v] == 0:
                self._queue.put(v)

    def _start_traverse(self):
        """Initialize graph traversing"""
        self._run_in_degree = self._get_in_degree()
        self._queue = queue.Queue()
        self._done = set()
        self._it_cnt = len(self._graph)

        for k, v in self._run_in_degree.items():
            if v == 0:
                self._queue.put(k)

    def _get_in_degree(self):
        """Calculate the in_degree (count incoming) for nodes"""
        _in_degree = dict()
        _in_degree = {u: 0 for u in self._graph.keys()}
        for u in self._graph:
            for v in self._graph[u]:
                _in_degree[v] += 1

        return _in_degree

    def topological_sort(self):
        """Return the graph nodes in the topological order"""
        result = []
        for node in self:
            result.append(node)
            self.node_done(node)

        return result

    def size(self):
        return len(self._graph.keys())

    def is_complete(self):
        return len(self._done) == self.size()



# Licensed under the Apache License, Version 2.0 (the "License"); you may

# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.object_store.v1 import _base
from openstack import resource


class Container(_base.BaseResource):
    _custom_metadata_prefix = "X-Container-Meta-"
    _system_metadata = {
        "content_type": "content-type",
        "is_content_type_detected": "x-detect-content-type",
        "versions_location": "x-versions-location",
        "history_location": "x-history-location",
        "read_ACL": "x-container-read",
        "write_ACL": "x-container-write",
        "sync_to": "x-container-sync-to",
        "sync_key": "x-container-sync-key"
    }

    base_path = "/"
    pagination_key = 'X-Account-Container-Count'

    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_head = True

    _query_mapping = resource.QueryParameters(
        'prefix', 'format'
    )

    # Container body data (when id=None)
    #: The name of the container.
    name = resource.Body("name", alternate_id=True, alias='id')
    #: The number of objects in the container.
    count = resource.Body("count", type=int, alias='object_count')
    #: The total number of bytes that are stored in Object Storage
    #: for the container.
    bytes = resource.Body("bytes", type=int, alias='bytes_used')

    # Container metadata (when id=name)
    #: The number of objects.
    object_count = resource.Header(
        "x-container-object-count", type=int, alias='count')
    #: The count of bytes used in total.
    bytes_used = resource.Header(
        "x-container-bytes-used", type=int, alias='bytes')
    #: The timestamp of the transaction.
    timestamp = resource.Header("x-timestamp")

    # Request headers (when id=None)
    #: If set to True, Object Storage queries all replicas to return the
    #: most recent one. If you omit this header, Object Storage responds
    #: faster after it finds one valid replica. Because setting this
    #: header to True is more expensive for the back end, use it only
    #: when it is absolutely needed. *Type: bool*
    is_newest = resource.Header("x-newest", type=bool)

    # Request headers (when id=name)
    #: The ACL that grants read access. If not set, this header is not
    #: returned by this operation.
    read_ACL = resource.Header("x-container-read")
    #: The ACL that grants write access. If not set, this header is not
    #: returned by this operation.
    write_ACL = resource.Header("x-container-write")
    #: The destination for container synchronization. If not set,
    #: this header is not returned by this operation.
    sync_to = resource.Header("x-container-sync-to")
    #: The secret key for container synchronization. If not set,
    #: this header is not returned by this operation.
    sync_key = resource.Header("x-container-sync-key")
    #: Enables versioning on this container. The value is the name
    #: of another container. You must UTF-8-encode and then URL-encode
    #: the name before you include it in the header. To disable
    #: versioning, set the header to an empty string.
    versions_location = resource.Header("x-versions-location")
    #: Enables versioning on the container.
    history_location = resource.Header("x-history-location")
    #: The MIME type of the list of names.
    content_type = resource.Header("content-type")
    #: If set to true, Object Storage guesses the content type based
    #: on the file extension and ignores the value sent in the
    #: Content-Type header, if present. *Type: bool*
    is_content_type_detected = resource.Header("x-detect-content-type",
                                               type=bool)

    #: Storage policy used by the container.
    #: It is not possible to change policy of an existing container
    storage_policy = resource.Header("x-storage-policy")
    # TODO(mordred) Shouldn't if-none-match be handled more systemically?
    #: In combination with Expect: 100-Continue, specify an
    #: "If-None-Match: \*" header to query whether the server already
    #: has a copy of the object before any data is sent.
    if_none_match = resource.Header("if-none-match")
    #: The secret key value for temporary URLs. If not set,
    #: this header is not returned by this operation.
    meta_temp_url_key = resource.Header("x-container-meta-temp-url-key")
    #: A second secret key value for temporary URLs. If not set,
    #: this header is not returned by this operation.
    meta_temp_url_key_2 = resource.Header("x-container-meta-temp-url-key-2")

    @classmethod
    def new(cls, **kwargs):
        # Container uses name as id. Proxy._get_resource calls
        # Resource.new(id=name) but then we need to do container.name
        # It's the same thing for Container - make it be the same.
        name = kwargs.pop('id', None)
        if name:
            kwargs.setdefault('name', name)
        return cls(_synchronized=False, **kwargs)

    def create(self, session, prepend_key=True, base_path=None):
        """Create a remote resource based on this instance.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param prepend_key: A boolean indicating whether the resource_key
                            should be prepended in a resource creation
                            request. Default to True.

        :return: This :class:`Resource` instance.
        :raises: :exc:`~openstack.exceptions.MethodNotSupported` if
                 :data:`Resource.allow_create` is not set to ``True``.
        """
        request = self._prepare_request(
            requires_id=True, prepend_key=prepend_key, base_path=base_path)
        response = session.put(
            request.url, headers=request.headers)

        self._translate_response(response, has_body=False)
        return self

    def set_temp_url_key(self, proxy, key, secondary=False):
        """Set the temporary url key for a container.

        :param proxy: The proxy to use for making this request.
        :type proxy: :class:`~openstack.proxy.Proxy`
        :param container:
          The value can be the name of a container or a
          :class:`~openstack.object_store.v1.container.Container` instance.
        :param key:
          Text of the key to use.
        :param bool secondary:
          Whether this should set the second key. (defaults to False)
        """
        header = 'Temp-URL-Key'
        if secondary:
            header += '-2'

        return self.set_metadata(proxy, {header: key})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class SoftwareConfig(resource.Resource):
    resource_key = 'software_config'
    resources_key = 'software_configs'
    base_path = '/software_configs'

    # capabilities
    allow_create = True
    allow_list = True
    allow_fetch = True
    allow_delete = True
    allow_commit = False

    # Properties
    #: Configuration script or manifest that defines which configuration is
    #: performed
    config = resource.Body('config')
    #: The date and time when the software config resource was created.
    created_at = resource.Body('creation_time')
    #: A string indicating the namespace used for grouping software configs.
    group = resource.Body('group')
    #: A list of schemas each representing an input this software config
    #: expects.
    inputs = resource.Body('inputs')
    #: Name of the software config.
    name = resource.Body('name')
    #: A string that contains options that are specific to the configuration
    #: management tool that this resource uses.
    options = resource.Body('options')
    #: A list of schemas each representing an output this software config
    #: produces.
    outputs = resource.Body('outputs')

    def create(self, session, base_path=None):
        # This overrides the default behavior of resource creation because
        # heat doesn't accept resource_key in its request.
        return super(SoftwareConfig, self).create(session, prepend_key=False,
                                                  base_path=base_path)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import availability_zone as az
from openstack.tests.unit import base

IDENTIFIER = '08a87d37-5ca2-4308-86c5-cba06d8d796c'
EXAMPLE = {
    "id": IDENTIFIER,
    "name": "nova",
    "created_at": "2021-01-21T20:13:55.000000",
    "updated_at": None,
}


class TestAvailabilityZone(base.TestCase):

    def test_basic(self):
        az_resource = az.AvailabilityZone()
        self.assertEqual('availability_zones', az_resource.resources_key)
        self.assertEqual('/availability-zones', az_resource.base_path)
        self.assertTrue(az_resource.allow_list)

    def test_make_availability_zone(self):
        az_resource = az.AvailabilityZone(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], az_resource.id)
        self.assertEqual(EXAMPLE['name'], az_resource.name)
        self.assertEqual(EXAMPLE['created_at'], az_resource.created_at)
        self.assertEqual(EXAMPLE['updated_at'], az_resource.updated_at)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import fnmatch
import functools
import inspect
import re
import time
import uuid

from decorator import decorator
import jmespath
import netifaces
import sre_constants

from openstack import _log
from openstack.cloud import exc


_decorated_methods = []


def _make_unicode(input):
    """Turn an input into unicode unconditionally

    :param input: A unicode, string or other object
    """
    try:
        if isinstance(input, unicode):
            return input
        if isinstance(input, str):
            return input.decode('utf-8')
        else:
            # int, for example
            return unicode(input)
    except NameError:
        # python3!
        return str(input)


def _dictify_resource(resource):
    if isinstance(resource, list):
        return [_dictify_resource(r) for r in resource]
    else:
        if hasattr(resource, 'toDict'):
            return resource.toDict()
        else:
            return resource


def _filter_list(data, name_or_id, filters):
    """Filter a list by name/ID and arbitrary meta data.

    :param list data: The list of dictionary data to filter. It is expected
        that each dictionary contains an 'id' and 'name' key if a value for
        name_or_id is given.
    :param string name_or_id: The name or ID of the entity being filtered. Can
        be a glob pattern, such as 'nb01*'.
    :param filters: A dictionary of meta data to use for further filtering.
        Elements of this dictionary may, themselves, be dictionaries. Example::

            {
                'last_name': 'Smith',
                'other': {
                    'gender': 'Female'
                }
            }

        OR

        A string containing a jmespath expression for further filtering.
    """
    # The logger is openstack.cloud.fmmatch to allow a user/operator to
    # configure logging not to communicate about fnmatch misses
    # (they shouldn't be too spammy, but one never knows)
    log = _log.setup_logging('openstack.fnmatch')
    if name_or_id:
        # name_or_id might already be unicode
        name_or_id = _make_unicode(name_or_id)
        identifier_matches = []
        bad_pattern = False
        try:
            fn_reg = re.compile(fnmatch.translate(name_or_id))
        except sre_constants.error:
            # If the fnmatch re doesn't compile, then we don't care,
            # but log it in case the user DID pass a pattern but did
            # it poorly and wants to know what went wrong with their
            # search
            fn_reg = None
        for e in data:
            e_id = _make_unicode(e.get('id', None))
            e_name = _make_unicode(e.get('name', None))

            if ((e_id and e_id == name_or_id)
                    or (e_name and e_name == name_or_id)):
                identifier_matches.append(e)
            else:
                # Only try fnmatch if we don't match exactly
                if not fn_reg:
                    # If we don't have a pattern, skip this, but set the flag
                    # so that we log the bad pattern
                    bad_pattern = True
                    continue
                if ((e_id and fn_reg.match(e_id))
                        or (e_name and fn_reg.match(e_name))):
                    identifier_matches.append(e)
        if not identifier_matches and bad_pattern:
            log.debug("Bad pattern passed to fnmatch", exc_info=True)
        data = identifier_matches

    if not filters:
        return data

    if isinstance(filters, str):
        return jmespath.search(filters, data)

    def _dict_filter(f, d):
        if not d:
            return False
        for key in f.keys():
            if isinstance(f[key], dict):
                if not _dict_filter(f[key], d.get(key, None)):
                    return False
            elif d.get(key, None) != f[key]:
                return False
        return True

    filtered = []
    for e in data:
        filtered.append(e)
        for key in filters.keys():
            if isinstance(filters[key], dict):
                if not _dict_filter(filters[key], e.get(key, None)):
                    filtered.pop()
                    break
            elif e.get(key, None) != filters[key]:
                filtered.pop()
                break
    return filtered


def _get_entity(cloud, resource, name_or_id, filters, **kwargs):
    """Return a single entity from the list returned by a given method.

    :param object cloud: The controller class (Example: the main OpenStackCloud
        object).
    :param string or callable resource: The string that identifies the resource
        to use to lookup the get_<>_by_id or search_<resource>s methods
        (Example: network) or a callable to invoke.
    :param string name_or_id: The name or ID of the entity being filtered or an
        object or dict. If this is an object/dict with an 'id' attr/key, we
        return it and bypass resource lookup.
    :param filters: A dictionary of meta data to use for further filtering.
        OR A string containing a jmespath expression for further filtering.
        Example:: "[?last_name==`Smith`] | [?other.gender]==`Female`]"
    """

    # Sometimes in the control flow of shade, we already have an object
    # fetched. Rather than then needing to pull the name or id out of that
    # object, pass it in here and rely on caching to prevent us from making
    # an additional call, it's simple enough to test to see if we got an
    # object and just short-circuit return it.

    if (hasattr(name_or_id, 'id')
            or (isinstance(name_or_id, dict) and 'id' in name_or_id)):
        return name_or_id

    # If a uuid is passed short-circuit it calling the
    # get_<resource_name>_by_id method
    if getattr(cloud, 'use_direct_get', False) and _is_uuid_like(name_or_id):
        get_resource = getattr(cloud, 'get_%s_by_id' % resource, None)
        if get_resource:
            return get_resource(name_or_id)

    search = resource if callable(resource) else getattr(
        cloud, 'search_%ss' % resource, None)
    if search:
        entities = search(name_or_id, filters, **kwargs)
        if entities:
            if len(entities) > 1:
                raise exc.OpenStackCloudException(
                    "Multiple matches found for %s" % name_or_id)
            return entities[0]
    return None


def localhost_supports_ipv6():
    """Determine whether the local host supports IPv6

    We look for a default route that supports the IPv6 address family,
    and assume that if it is present, this host has globally routable
    IPv6 connectivity.
    """

    try:
        return netifaces.AF_INET6 in netifaces.gateways()['default']
    except AttributeError:
        return False


def valid_kwargs(*valid_args):
    # This decorator checks if argument passed as **kwargs to a function are
    # present in valid_args.
    #
    # Typically, valid_kwargs is used when we want to distinguish between
    # None and omitted arguments and we still want to validate the argument
    # list.
    #
    # Example usage:
    #
    # @valid_kwargs('opt_arg1', 'opt_arg2')
    # def my_func(self, mandatory_arg1, mandatory_arg2, **kwargs):
    #   ...
    #
    @decorator
    def func_wrapper(func, *args, **kwargs):
        argspec = inspect.getfullargspec(func)
        for k in kwargs:
            if k not in argspec.args[1:] and k not in valid_args:
                raise TypeError(
                    "{f}() got an unexpected keyword argument "
                    "'{arg}'".format(f=inspect.stack()[1][3], arg=k))
        return func(*args, **kwargs)
    return func_wrapper


def _func_wrap(f):
    # NOTE(morgan): This extra wrapper is intended to eliminate ever
    # passing a bound method to dogpile.cache's cache_on_arguments. In
    # 0.7.0 and later it is impossible to pass bound methods to the
    # decorator. This was introduced when utilizing the decorate module in
    # lieu of a direct wrap implementation.
    @functools.wraps(f)
    def inner(*args, **kwargs):
        return f(*args, **kwargs)
    return inner


def cache_on_arguments(*cache_on_args, **cache_on_kwargs):
    _cache_name = cache_on_kwargs.pop('resource', None)

    def _inner_cache_on_arguments(func):
        def _cache_decorator(obj, *args, **kwargs):
            the_method = obj._get_cache(_cache_name).cache_on_arguments(
                *cache_on_args, **cache_on_kwargs)(
                    _func_wrap(func.__get__(obj, type(obj))))
            return the_method(*args, **kwargs)

        def invalidate(obj, *args, **kwargs):
            return obj._get_cache(
                _cache_name).cache_on_arguments()(func).invalidate(
                    *args, **kwargs)

        _cache_decorator.invalidate = invalidate
        _cache_decorator.func = func
        _decorated_methods.append(func.__name__)

        return _cache_decorator
    return _inner_cache_on_arguments


@contextlib.contextmanager
def shade_exceptions(error_message=None):
    """Context manager for dealing with shade exceptions.

    :param string error_message: String to use for the exception message
        content on non-OpenStackCloudExceptions.

        Useful for avoiding wrapping shade OpenStackCloudException exceptions
        within themselves. Code called from within the context may throw such
        exceptions without having to catch and reraise them.

        Non-OpenStackCloudException exceptions thrown within the context will
        be wrapped and the exception message will be appended to the given
        error message.
    """
    try:
        yield
    except exc.OpenStackCloudException:
        raise
    except Exception as e:
        if error_message is None:
            error_message = str(e)
        raise exc.OpenStackCloudException(error_message)


def safe_dict_min(key, data):
    """Safely find the minimum for a given key in a list of dict objects.

    This will find the minimum integer value for specific dictionary key
    across a list of dictionaries. The values for the given key MUST be
    integers, or string representations of an integer.

    The dictionary key does not have to be present in all (or any)
    of the elements/dicts within the data set.

    :param string key: The dictionary key to search for the minimum value.
    :param list data: List of dicts to use for the data set.

    :returns: None if the field was not found in any elements, or
        the minimum value for the field otherwise.
    """
    min_value = None
    for d in data:
        if (key in d) and (d[key] is not None):
            try:
                val = int(d[key])
            except ValueError:
                raise exc.OpenStackCloudException(
                    "Search for minimum value failed. "
                    "Value for {key} is not an integer: {value}".format(
                        key=key, value=d[key])
                )
            if (min_value is None) or (val < min_value):
                min_value = val
    return min_value


def safe_dict_max(key, data):
    """Safely find the maximum for a given key in a list of dict objects.

    This will find the maximum integer value for specific dictionary key
    across a list of dictionaries. The values for the given key MUST be
    integers, or string representations of an integer.

    The dictionary key does not have to be present in all (or any)
    of the elements/dicts within the data set.

    :param string key: The dictionary key to search for the maximum value.
    :param list data: List of dicts to use for the data set.

    :returns: None if the field was not found in any elements, or
        the maximum value for the field otherwise.
    """
    max_value = None
    for d in data:
        if (key in d) and (d[key] is not None):
            try:
                val = int(d[key])
            except ValueError:
                raise exc.OpenStackCloudException(
                    "Search for maximum value failed. "
                    "Value for {key} is not an integer: {value}".format(
                        key=key, value=d[key])
                )
            if (max_value is None) or (val > max_value):
                max_value = val
    return max_value


def _call_client_and_retry(client, url, retry_on=None,
                           call_retries=3, retry_wait=2,
                           **kwargs):
    """Method to provide retry operations.

    Some APIs utilize HTTP errors on certain operations to indicate that
    the resource is presently locked, and as such this mechanism provides
    the ability to retry upon known error codes.

    :param object client: The client method, such as:
        ``self.baremetal_client.post``
    :param string url: The URL to perform the operation upon.
    :param integer retry_on: A list of error codes that can be retried on.
        The method also supports a single integer to be
        defined.
    :param integer call_retries: The number of times to retry the call upon
        the error code defined by the 'retry_on' parameter. Default: 3
    :param integer retry_wait: The time in seconds to wait between retry
        attempts. Default: 2

    :returns: The object returned by the client call.
    """

    # NOTE(TheJulia): This method, as of this note, does not have direct
    # unit tests, although is fairly well tested by the tests checking
    # retry logic in test_baremetal_node.py.
    log = _log.setup_logging('shade.http')

    if isinstance(retry_on, int):
        retry_on = [retry_on]

    count = 0
    while (count < call_retries):
        count += 1
        try:
            ret_val = client(url, **kwargs)
        except exc.OpenStackCloudHTTPError as e:
            if (retry_on is not None
                    and e.response.status_code in retry_on):
                log.debug('Received retryable error %(err)s, waiting '
                          '%(wait)s seconds to retry', {
                              'err': e.response.status_code,
                              'wait': retry_wait
                          })
                time.sleep(retry_wait)
                continue
            else:
                raise
        # Break out of the loop, since the loop should only continue
        # when we encounter a known connection error.
        return ret_val


def parse_range(value):
    """Parse a numerical range string.

    Breakdown a range expression into its operater and numerical parts.
    This expression must be a string. Valid values must be an integer string,
    optionally preceeded by one of the following operators::

        - "<"  : Less than
        - ">"  : Greater than
        - "<=" : Less than or equal to
        - ">=" : Greater than or equal to

    Some examples of valid values and function return values::

        - "1024"  : returns (None, 1024)
        - "<5"    : returns ("<", 5)
        - ">=100" : returns (">=", 100)

    :param string value: The range expression to be parsed.

    :returns: A tuple with the operator string (or None if no operator
        was given) and the integer value. None is returned if parsing failed.
    """
    if value is None:
        return None

    range_exp = re.match(r'(<|>|<=|>=){0,1}(\d+)$', value)
    if range_exp is None:
        return None

    op = range_exp.group(1)
    num = int(range_exp.group(2))
    return (op, num)


def range_filter(data, key, range_exp):
    """Filter a list by a single range expression.

    :param list data: List of dictionaries to be searched.
    :param string key: Key name to search within the data set.
    :param string range_exp: The expression describing the range of values.

    :returns: A list subset of the original data set.
    :raises: OpenStackCloudException on invalid range expressions.
    """
    filtered = []
    range_exp = str(range_exp).upper()

    if range_exp == "MIN":
        key_min = safe_dict_min(key, data)
        if key_min is None:
            return []
        for d in data:
            if int(d[key]) == key_min:
                filtered.append(d)
        return filtered
    elif range_exp == "MAX":
        key_max = safe_dict_max(key, data)
        if key_max is None:
            return []
        for d in data:
            if int(d[key]) == key_max:
                filtered.append(d)
        return filtered

    # Not looking for a min or max, so a range or exact value must
    # have been supplied.
    val_range = parse_range(range_exp)

    # If parsing the range fails, it must be a bad value.
    if val_range is None:
        raise exc.OpenStackCloudException(
            "Invalid range value: {value}".format(value=range_exp))

    op = val_range[0]
    if op:
        # Range matching
        for d in data:
            d_val = int(d[key])
            if op == '<':
                if d_val < val_range[1]:
                    filtered.append(d)
            elif op == '>':
                if d_val > val_range[1]:
                    filtered.append(d)
            elif op == '<=':
                if d_val <= val_range[1]:
                    filtered.append(d)
            elif op == '>=':
                if d_val >= val_range[1]:
                    filtered.append(d)
        return filtered
    else:
        # Exact number match
        for d in data:
            if int(d[key]) == val_range[1]:
                filtered.append(d)
        return filtered


def generate_patches_from_kwargs(operation, **kwargs):
    """Given a set of parameters, returns a list with the
    valid patch values.

    :param string operation: The operation to perform.
    :param list kwargs: Dict of parameters.

    :returns: A list with the right patch values.
    """
    patches = []
    for k, v in kwargs.items():
        patch = {'op': operation,
                 'value': v,
                 'path': '/%s' % k}
        patches.append(patch)
    return sorted(patches)


class FileSegment:
    """File-like object to pass to requests."""

    def __init__(self, filename, offset, length):
        self.filename = filename
        self.offset = offset
        self.length = length
        self.pos = 0
        self._file = open(filename, 'rb')
        self.seek(0)

    def tell(self):
        return self._file.tell() - self.offset

    def seek(self, offset, whence=0):
        if whence == 0:
            self._file.seek(self.offset + offset, whence)
        elif whence == 1:
            self._file.seek(offset, whence)
        elif whence == 2:
            self._file.seek(self.offset + self.length - offset, 0)

    def read(self, size=-1):
        remaining = self.length - self.pos
        if remaining <= 0:
            return b''

        to_read = remaining if size < 0 else min(size, remaining)
        chunk = self._file.read(to_read)
        self.pos += len(chunk)

        return chunk

    def reset(self):
        self._file.seek(self.offset, 0)


def _format_uuid_string(string):
    return (string.replace('urn:', '')
                  .replace('uuid:', '')
                  .strip('{}')
                  .replace('-', '')
                  .lower())


def _is_uuid_like(val):
    """Returns validation of a value as a UUID.

    :param val: Value to verify
    :type val: string
    :returns: bool

    .. versionchanged:: 1.1.1
       Support non-lowercase UUIDs.
    """
    try:
        return str(uuid.UUID(val)).replace('-', '') == _format_uuid_string(val)
    except (TypeError, ValueError, AttributeError):
        return False



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


def normalize_keys(config):
    new_config = {}
    for key, value in config.items():
        key = key.replace('-', '_')
        if isinstance(value, dict):
            new_config[key] = normalize_keys(value)
        elif isinstance(value, bool):
            new_config[key] = value
        elif isinstance(value, int) and key not in (
                'verbose_level', 'api_timeout'):
            new_config[key] = str(value)
        elif isinstance(value, float):
            new_config[key] = str(value)
        else:
            new_config[key] = value
    return new_config


def merge_clouds(old_dict, new_dict):
    """Like dict.update, except handling nested dicts."""
    ret = old_dict.copy()
    for (k, v) in new_dict.items():
        if isinstance(v, dict):
            if k in ret:
                ret[k] = merge_clouds(ret[k], v)
            else:
                ret[k] = v.copy()
        else:
            ret[k] = v
    return ret


class VersionRequest:
    def __init__(
        self,
        version=None,
        min_api_version=None,
        max_api_version=None,
        default_microversion=None,
    ):
        self.version = version
        self.min_api_version = min_api_version
        self.max_api_version = max_api_version
        self.default_microversion = default_microversion



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.tests.functional.block_storage.v3 import base


class TestLimits(base.BaseBlockStorageTest):

    def test_get(self):
        sot = self.conn.block_storage.get_limits()
        self.assertIsNotNone(sot.absolute.max_total_backup_gigabytes)
        self.assertIsNotNone(sot.absolute.max_total_backups)
        self.assertIsNotNone(sot.absolute.max_total_snapshots)
        self.assertIsNotNone(sot.absolute.max_total_volume_gigabytes)
        self.assertIsNotNone(sot.absolute.max_total_volumes)
        self.assertIsNotNone(sot.absolute.total_backup_gigabytes_used)
        self.assertIsNotNone(sot.absolute.total_backups_used)
        self.assertIsNotNone(sot.absolute.total_gigabytes_used)
        self.assertIsNotNone(sot.absolute.total_snapshots_used)
        self.assertIsNotNone(sot.absolute.total_volumes_used)
        self.assertIsNotNone(sot.rate)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import json
import os

import jsonschema
from testtools import content

from openstack.config import defaults
from openstack.tests.unit.config import base


class TestConfig(base.TestCase):

    def json_diagnostics(self, exc_info):
        self.addDetail('filename', content.text_content(self.filename))
        for error in sorted(self.validator.iter_errors(self.json_data)):
            self.addDetail('jsonschema', content.text_content(str(error)))

    def test_defaults_valid_json(self):
        _schema_path = os.path.join(
            os.path.dirname(os.path.realpath(defaults.__file__)),
            'schema.json')
        with open(_schema_path, 'r') as f:
            schema = json.load(f)
        self.validator = jsonschema.Draft4Validator(schema)
        self.addOnException(self.json_diagnostics)

        self.filename = os.path.join(
            os.path.dirname(os.path.realpath(defaults.__file__)),
            'defaults.json')
        with open(self.filename, 'r') as f:
            self.json_data = json.load(f)

        self.assertTrue(self.validator.is_valid(self.json_data))

    def test_vendors_valid_json(self):
        _schema_path = os.path.join(
            os.path.dirname(os.path.realpath(defaults.__file__)),
            'vendor-schema.json')
        with open(_schema_path, 'r') as f:
            schema = json.load(f)
            self.validator = jsonschema.Draft4Validator(schema)

        self.addOnException(self.json_diagnostics)

        _vendors_path = os.path.join(
            os.path.dirname(os.path.realpath(defaults.__file__)),
            'vendors')
        for self.filename in glob.glob(os.path.join(_vendors_path, '*.json')):
            with open(self.filename, 'r') as f:
                self.json_data = json.load(f)

                self.assertTrue(self.validator.is_valid(self.json_data))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ServiceProvider(resource.Resource):
    resources_key = 'service_providers'
    base_path = '/service-providers'

    _allow_unknown_attrs_in_body = True

    # Capabilities
    allow_create = False
    allow_fetch = False
    allow_commit = False
    allow_delete = False
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'service_type', 'name',
        is_default='default'
    )

    # Properties
    #: Service type (FIREWALL, FLAVORS, METERING, QOS, etc..)
    service_type = resource.Body('service_type')
    #: Name of the service type
    name = resource.Body('name')
    #: The default value of service type
    is_default = resource.Body('default', type=bool)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import cloud as openstack

# Initialize and turn on debug logging
openstack.enable_logging(debug=True)

for cloud_name, region_name in [
        ('my-vexxhost', 'ca-ymq-1'),
        ('my-citycloud', 'Buf1'),
        ('my-internap', 'ams01')]:
    # Initialize cloud
    cloud = openstack.connect(
        cloud=cloud_name, region_name=region_name)
    for server in cloud.search_servers('my-server'):
        cloud.delete_server(server, wait=True, delete_ips=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import uuid

from openstack.load_balancer.v2 import load_balancer
from openstack.tests.unit import base

IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'admin_state_up': True,
    'availability_zone': 'my_fake_az',
    'created_at': '2017-07-17T12:14:57.233772',
    'description': 'fake_description',
    'flavor_id': uuid.uuid4(),
    'id': IDENTIFIER,
    'listeners': [{'id', uuid.uuid4()}],
    'name': 'test_load_balancer',
    'operating_status': 'ONLINE',
    'pools': [{'id', uuid.uuid4()}],
    'project_id': uuid.uuid4(),
    'provider': 'fake_provider',
    'provisioning_status': 'ACTIVE',
    'updated_at': '2017-07-17T12:16:57.233772',
    'vip_address': '192.0.2.5',
    'vip_network_id': uuid.uuid4(),
    'vip_port_id': uuid.uuid4(),
    'vip_subnet_id': uuid.uuid4(),
    'vip_qos_policy_id': uuid.uuid4(),
}

EXAMPLE_STATS = {
    'active_connections': 1,
    'bytes_in': 2,
    'bytes_out': 3,
    'request_errors': 4,
    'total_connections': 5
}


class TestLoadBalancer(base.TestCase):

    def test_basic(self):
        test_load_balancer = load_balancer.LoadBalancer()
        self.assertEqual('loadbalancer', test_load_balancer.resource_key)
        self.assertEqual('loadbalancers', test_load_balancer.resources_key)
        self.assertEqual('/lbaas/loadbalancers',
                         test_load_balancer.base_path)
        self.assertTrue(test_load_balancer.allow_create)
        self.assertTrue(test_load_balancer.allow_fetch)
        self.assertTrue(test_load_balancer.allow_delete)
        self.assertTrue(test_load_balancer.allow_list)
        self.assertTrue(test_load_balancer.allow_commit)

    def test_make_it(self):
        test_load_balancer = load_balancer.LoadBalancer(**EXAMPLE)
        self.assertTrue(test_load_balancer.is_admin_state_up)
        self.assertEqual(EXAMPLE['availability_zone'],
                         test_load_balancer.availability_zone)
        self.assertEqual(EXAMPLE['created_at'], test_load_balancer.created_at)
        self.assertEqual(EXAMPLE['description'],
                         test_load_balancer.description)
        self.assertEqual(EXAMPLE['flavor_id'], test_load_balancer.flavor_id)
        self.assertEqual(EXAMPLE['id'], test_load_balancer.id)
        self.assertEqual(EXAMPLE['listeners'], test_load_balancer.listeners)
        self.assertEqual(EXAMPLE['name'], test_load_balancer.name)
        self.assertEqual(EXAMPLE['operating_status'],
                         test_load_balancer.operating_status)
        self.assertEqual(EXAMPLE['pools'], test_load_balancer.pools)
        self.assertEqual(EXAMPLE['project_id'], test_load_balancer.project_id)
        self.assertEqual(EXAMPLE['provider'], test_load_balancer.provider)
        self.assertEqual(EXAMPLE['provisioning_status'],
                         test_load_balancer.provisioning_status)
        self.assertEqual(EXAMPLE['updated_at'], test_load_balancer.updated_at)
        self.assertEqual(EXAMPLE['vip_address'],
                         test_load_balancer.vip_address)
        self.assertEqual(EXAMPLE['vip_network_id'],
                         test_load_balancer.vip_network_id)
        self.assertEqual(EXAMPLE['vip_port_id'],
                         test_load_balancer.vip_port_id)
        self.assertEqual(EXAMPLE['vip_subnet_id'],
                         test_load_balancer.vip_subnet_id)
        self.assertEqual(EXAMPLE['vip_qos_policy_id'],
                         test_load_balancer.vip_qos_policy_id)

        self.assertDictEqual(
            {'limit': 'limit',
             'marker': 'marker',
             'availability_zone': 'availability_zone',
             'description': 'description',
             'flavor_id': 'flavor_id',
             'name': 'name',
             'project_id': 'project_id',
             'provider': 'provider',
             'operating_status': 'operating_status',
             'provisioning_status': 'provisioning_status',
             'is_admin_state_up': 'admin_state_up',
             'vip_address': 'vip_address',
             'vip_network_id': 'vip_network_id',
             'vip_port_id': 'vip_port_id',
             'vip_subnet_id': 'vip_subnet_id',
             'vip_qos_policy_id': 'vip_qos_policy_id',
             'tags': 'tags',
             'any_tags': 'tags-any',
             'not_tags': 'not-tags',
             'not_any_tags': 'not-tags-any',
             },
            test_load_balancer._query_mapping._mapping)

    def test_delete_non_cascade(self):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.delete.return_value = resp

        sot = load_balancer.LoadBalancer(**EXAMPLE)
        sot.cascade = False
        sot._translate_response = mock.Mock()
        sot.delete(sess)

        url = 'lbaas/loadbalancers/%(lb)s' % {
            'lb': EXAMPLE['id']
        }
        headers = {'Accept': ''}
        params = {}
        sess.delete.assert_called_with(url,
                                       headers=headers,
                                       params=params)
        sot._translate_response.assert_called_once_with(
            resp,
            error_message=None,
            has_body=False,
        )

    def test_delete_cascade(self):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.delete.return_value = resp

        sot = load_balancer.LoadBalancer(**EXAMPLE)
        sot.cascade = True
        sot._translate_response = mock.Mock()
        sot.delete(sess)

        url = 'lbaas/loadbalancers/%(lb)s' % {
            'lb': EXAMPLE['id']
        }
        headers = {'Accept': ''}
        params = {'cascade': True}
        sess.delete.assert_called_with(url,
                                       headers=headers,
                                       params=params)
        sot._translate_response.assert_called_once_with(
            resp,
            error_message=None,
            has_body=False,
        )


class TestLoadBalancerStats(base.TestCase):

    def test_basic(self):
        test_load_balancer = load_balancer.LoadBalancerStats()
        self.assertEqual('stats', test_load_balancer.resource_key)
        self.assertEqual('/lbaas/loadbalancers/%(lb_id)s/stats',
                         test_load_balancer.base_path)
        self.assertFalse(test_load_balancer.allow_create)
        self.assertTrue(test_load_balancer.allow_fetch)
        self.assertFalse(test_load_balancer.allow_delete)
        self.assertFalse(test_load_balancer.allow_list)
        self.assertFalse(test_load_balancer.allow_commit)

    def test_make_it(self):
        test_load_balancer = load_balancer.LoadBalancerStats(**EXAMPLE_STATS)
        self.assertEqual(EXAMPLE_STATS['active_connections'],
                         test_load_balancer.active_connections)
        self.assertEqual(EXAMPLE_STATS['bytes_in'],
                         test_load_balancer.bytes_in)
        self.assertEqual(EXAMPLE_STATS['bytes_out'],
                         test_load_balancer.bytes_out)
        self.assertEqual(EXAMPLE_STATS['request_errors'],
                         test_load_balancer.request_errors)
        self.assertEqual(EXAMPLE_STATS['total_connections'],
                         test_load_balancer.total_connections)


class TestLoadBalancerFailover(base.TestCase):

    def test_basic(self):
        test_load_balancer = load_balancer.LoadBalancerFailover()
        self.assertEqual('/lbaas/loadbalancers/%(lb_id)s/failover',
                         test_load_balancer.base_path)
        self.assertFalse(test_load_balancer.allow_create)
        self.assertFalse(test_load_balancer.allow_fetch)
        self.assertFalse(test_load_balancer.allow_delete)
        self.assertFalse(test_load_balancer.allow_list)
        self.assertTrue(test_load_balancer.allow_commit)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack import resource


class AcceleratorRequest(resource.Resource):
    resource_key = 'arq'
    resources_key = 'arqs'
    base_path = '/accelerator_requests'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_delete = True
    allow_list = True
    #: Allow patch operation for binding.
    allow_patch = True

    #: The device address associated with this ARQ (if any)
    attach_handle_info = resource.Body('attach_handle_info')
    #: The type of attach handle (e.g. PCI, mdev...)
    attach_handle_type = resource.Body('attach_handle_type')
    #: The name of the device profile
    device_profile_name = resource.Body('device_profile_name')
    #: The id of the device profile group
    device_profile_group_id = resource.Body('device_profile_group_id')
    #: The UUID of the bound device RP (if any)
    device_rp_uuid = resource.Body('device_rp_uuid')
    #: The host name to which ARQ is bound. (if any)
    hostname = resource.Body('hostname')
    #: The UUID of the instance associated with this ARQ (if any)
    instance_uuid = resource.Body('instance_uuid')
    #: The state of the ARQ
    state = resource.Body('state')
    #: The UUID of the ARQ
    uuid = resource.Body('uuid', alternate_id=True)

    def _convert_patch(self, patch):
        # This overrides the default behavior of _convert_patch because
        # the PATCH method consumes JSON, its key is the ARQ uuid
        # and its value is an ordinary JSON patch. spec:
        # https://specs.openstack.org/openstack/cyborg-specs/specs/train/implemented/cyborg-api

        converted = super(AcceleratorRequest, self)._convert_patch(patch)
        converted = {self.id: converted}
        return converted

    def patch(self, session, patch=None, prepend_key=True, has_body=True,
              retry_on_conflict=None, base_path=None):
        # This overrides the default behavior of patch because
        # the PATCH method consumes a dict rather than a list. spec:
        # https://specs.openstack.org/openstack/cyborg-specs/specs/train/implemented/cyborg-api

        # The id cannot be dirty for an commit
        self._body._dirty.discard("id")

        # Only try to update if we actually have anything to commit.
        if not patch and not self.requires_commit:
            return self

        if not self.allow_patch:
            raise exceptions.MethodNotSupported(self, "patch")

        request = self._prepare_request(prepend_key=prepend_key,
                                        base_path=base_path, patch=True)
        microversion = self._get_microversion(session, action='patch')
        if patch:
            request.body = self._convert_patch(patch)

        return self._commit(session, request, 'PATCH', microversion,
                            has_body=has_body,
                            retry_on_conflict=retry_on_conflict)

    def _consume_attrs(self, mapping, attrs):
        # This overrides the default behavior of _consume_attrs because
        # cyborg api returns an ARQ as list. spec:
        # https://specs.openstack.org/openstack/cyborg-specs/specs/train/implemented/cyborg-api
        if isinstance(self, AcceleratorRequest):
            if self.resources_key in attrs:
                attrs = attrs[self.resources_key][0]
        return super(AcceleratorRequest, self)._consume_attrs(mapping, attrs)

    def create(self, session, base_path=None):
        # This overrides the default behavior of resource creation because
        # cyborg doesn't accept resource_key in its request.
        return super(AcceleratorRequest, self).create(
            session, prepend_key=False, base_path=base_path)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack.common import tag
from openstack import resource


class LoadBalancer(resource.Resource, tag.TagMixin):
    resource_key = 'loadbalancer'
    resources_key = 'loadbalancers'
    base_path = '/lbaas/loadbalancers'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'description', 'flavor_id', 'name', 'project_id', 'provider',
        'vip_address', 'vip_network_id', 'vip_port_id', 'vip_subnet_id',
        'vip_qos_policy_id', 'provisioning_status', 'operating_status',
        'availability_zone', is_admin_state_up='admin_state_up',
        **tag.TagMixin._tag_query_parameters
    )

    # Properties
    #: The administrative state of the load balancer *Type: bool*
    is_admin_state_up = resource.Body('admin_state_up', type=bool)
    #: Name of the target Octavia availability zone
    availability_zone = resource.Body('availability_zone')
    #: Timestamp when the load balancer was created
    created_at = resource.Body('created_at')
    #: The load balancer description
    description = resource.Body('description')
    #: The load balancer flavor ID
    flavor_id = resource.Body('flavor_id')
    #: List of listeners associated with this load balancer
    listeners = resource.Body('listeners', type=list)
    #: The load balancer name
    name = resource.Body('name')
    #: Operating status of the load balancer
    operating_status = resource.Body('operating_status')
    #: List of pools associated with this load balancer
    pools = resource.Body('pools', type=list)
    #: The ID of the project this load balancer is associated with.
    project_id = resource.Body('project_id')
    #: Provider name for the load balancer.
    provider = resource.Body('provider')
    #: The provisioning status of this load balancer
    provisioning_status = resource.Body('provisioning_status')
    #: Timestamp when the load balancer was last updated
    updated_at = resource.Body('updated_at')
    #: VIP address of load balancer
    vip_address = resource.Body('vip_address')
    #: VIP netowrk ID
    vip_network_id = resource.Body('vip_network_id')
    #: VIP port ID
    vip_port_id = resource.Body('vip_port_id')
    #: VIP subnet ID
    vip_subnet_id = resource.Body('vip_subnet_id')
    # VIP qos policy id
    vip_qos_policy_id = resource.Body('vip_qos_policy_id')

    def delete(self, session, error_message=None):
        request = self._prepare_request()
        headers = {
            "Accept": ""
        }

        request.headers.update(headers)
        params = {}
        if (hasattr(self, 'cascade') and isinstance(self.cascade, bool)
                and self.cascade):
            params['cascade'] = True
        response = session.delete(request.url,
                                  headers=headers,
                                  params=params)

        self._translate_response(response, has_body=False,
                                 error_message=error_message)
        return self


class LoadBalancerStats(resource.Resource):
    resource_key = 'stats'
    base_path = '/lbaas/loadbalancers/%(lb_id)s/stats'

    # capabilities
    allow_create = False
    allow_fetch = True
    allow_commit = False
    allow_delete = False
    allow_list = False

    # Properties
    #: The ID of the load balancer.
    lb_id = resource.URI('lb_id')
    #: The currently active connections.
    active_connections = resource.Body('active_connections', type=int)
    #: The total bytes received.
    bytes_in = resource.Body('bytes_in', type=int)
    #: The total bytes sent.
    bytes_out = resource.Body('bytes_out', type=int)
    #: The total requests that were unable to be fulfilled.
    request_errors = resource.Body('request_errors', type=int)
    #: The total connections handled.
    total_connections = resource.Body('total_connections', type=int)


class LoadBalancerFailover(resource.Resource):
    base_path = '/lbaas/loadbalancers/%(lb_id)s/failover'

    # capabilities
    allow_create = False
    allow_fetch = False
    allow_commit = True
    allow_delete = False
    allow_list = False
    allow_empty_commit = True

    requires_id = False

    # Properties
    #: The ID of the load balancer.
    lb_id = resource.URI('lb_id')

    # The default _update code path also has no
    # way to pass has_body into this function, so overriding the method here.
    def commit(self, session, base_path=None):
        return super(LoadBalancerFailover, self).commit(
            session, base_path=base_path, has_body=False)
